#include "calc.hpp"
#include "chess.hpp"
#include "config.hpp"
#include "eval.hpp"
#include "eval_values.hpp"
#include "fen.hpp"
#include "moves.hpp"
#include "pawn_structure_hash_table.hpp"
#include "selftest.hpp"
#include "string.hpp"
#include "util.hpp"
#include "zobrist.hpp"

#include <algorithm>
#include <fstream>
#include <iomanip>
#include <sstream>
#include <iostream>
#include <vector>


struct perft_ctx {
	perft_ctx()
		: move_ptr(moves)
	{
	}

	move_info moves[200 * (MAX_DEPTH) ];
	move_info* move_ptr;

	killer_moves killers;
};


void perft( perft_ctx& ctx, int depth, position const& p, color::type c, uint64_t& n )
{
	if( conf.depth == -1 ) {
		conf.depth = 8;
	}

	move_info* moves = ctx.move_ptr;

	check_map check( p, c );
	calculate_moves( p, c, ctx.move_ptr, check );

	if( !--depth ) {
		n += ctx.move_ptr - moves;
		ctx.move_ptr = moves;
		return;
	}

	for( move_info* it = moves; it != ctx.move_ptr; ++it ) {
		position new_pos = p;
		apply_move( new_pos, it->m, c );
		perft( ctx, depth, new_pos, static_cast<color::type>(1-c), n );
	}
	ctx.move_ptr = moves;
}

bool perft( std::size_t max_depth )
{
	uint64_t const perft_results[] = {
		20ull,
		400ull,
		8902ull,
		197281ull,
		4865609ull,
		119060324ull,
		3195901860ull,
		84998978956ull
	};

	/*
		alternate:r3k2r/Pppp1ppp/1b3nbN/nP6/BBP1P3/q4N2/Pp1P2PP/R2Q1RK1 w kq -
		6
		264
		9467
		422333
		15833292
		706045033
	*/

	perft_ctx ctx;
	for( unsigned int i = 0; i < (std::min)(max_depth, sizeof(perft_results)/sizeof(uint64_t)); ++i ) {
		ctx.move_ptr = ctx.moves;

		std::cerr << "Calculating number of possible moves in " << (i + 1) << " plies:" << std::endl;

		position p;
		init_board( p );
		color::type c = color::white;

		uint64_t ret = 0;

		int max_depth = i + 1;

		timestamp start;
		perft( ctx, max_depth, p, c, ret );
		timestamp stop;


		std::cerr << "Moves: "     << ret << std::endl;
		std::cerr << "Took:  "     << (stop - start).milliseconds() << " ms" << std::endl;
		if( ret ) {
			// Will overflow after ~3 months
			int64_t picoseconds = (stop-start).picoseconds();
			picoseconds /= ret;

			std::stringstream ss;
			ss << "Time/move: " << picoseconds / 1000 << "." << std::setw(1) << std::setfill('0') << (picoseconds / 100) % 10 << " ns" << std::endl;
			std::cerr << ss.str();
		}

		if( ret != perft_results[i] ) {
			std::cerr << "FAIL! Expected " << perft_results[i] << " moves." << std::endl;
			return false;
		}
		else {
			std::cerr << "PASS" << std::endl;
		}
		std::cerr << std::endl;
	}


	return true;
}


namespace {

static bool test_move_generation( std::string const& fen, std::string const& ref_moves )
{
	position p;
	color::type c;
	std::string error;
	if( !parse_fen_noclock( fen, p, c, &error ) ) {
		std::cerr << "Could not parse fen: " << error << std::endl;
		std::cerr << "Fen: " << fen << std::endl;
		return false;
	}

	move_info moves[200];
	move_info* pm = moves;
	check_map check( p, c );
	calculate_moves( p, c, pm, check );

	std::vector<std::string> ms;
	for( move_info* it = moves; it != pm; ++it ) {
		ms.push_back( move_to_string( it->m ) );
	}
	std::sort( ms.begin(), ms.end() );

	std::string s;
	for( std::size_t i = 0; i < ms.size(); ++i ) {
		if( i ) {
			s += " ";
		}
		std::size_t pos;
		while( (pos = ms[i].find( ' ' ) ) != std::string::npos ) {
			ms[i].erase( pos, 1 );
		}
		s += ms[i];
	}

	if( ref_moves != s ) {
		std::cerr << "Move mismatch!" << std::endl;
		std::cerr << "Fen: " << fen << std::endl;
		std::cerr << "Reference: " << ref_moves << std::endl;
		std::cerr << "Actual:    " << s << std::endl;

		return false;
	}

	return true;
}


static bool test_move_generation()
{
	if( !test_move_generation( "4R3/p1pp1p1p/b1n1rn2/1p2p1pP/1B1QPBr1/qPPb2PN/Pk1PNP2/R3K3 w Q g6",
						"c3-c4 f2-f3 h5-h6 h5xg6 Bb4-a5 Bb4-c5 Bb4-d6 Bb4-e7 Bb4-f8 Bb4xa3 Bf4-e3 Bf4xe5 Bf4xg5 Ke1-d1 Ke1-f1 Ne2-c1 Ne2-g1 Nh3-g1 Nh3xg5 Qd4-b6 Qd4-c4 Qd4-c5 Qd4-d5 Qd4-d6 Qd4-e3 Qd4xa7 Qd4xd3 Qd4xd7 Qd4xe5 Ra1-b1 Ra1-c1 Ra1-d1 Re8-a8 Re8-b8 Re8-c8 Re8-d8 Re8-e7 Re8-f8 Re8-g8 Re8-h8 Re8xe6" ) )
	{
		return false;
	}

	if( !test_move_generation( "r3k2r/8/2p5/1Q6/1q6/2P5/8/R3K2R w KqQk -",
						"O-O O-O-O c3xb4 Ke1-d1 Ke1-d2 Ke1-e2 Ke1-f1 Ke1-f2 Qb5-a4 Qb5-a5 Qb5-a6 Qb5-b6 Qb5-b7 Qb5-b8 Qb5-c4 Qb5-c5 Qb5-d3 Qb5-d5 Qb5-e2 Qb5-e5 Qb5-f1 Qb5-f5 Qb5-g5 Qb5-h5 Qb5xb4 Qb5xc6 Ra1-a2 Ra1-a3 Ra1-a4 Ra1-a5 Ra1-a6 Ra1-a7 Ra1-b1 Ra1-c1 Ra1-d1 Ra1xa8 Rh1-f1 Rh1-g1 Rh1-h2 Rh1-h3 Rh1-h4 Rh1-h5 Rh1-h6 Rh1-h7 Rh1xh8" ) )
	{
		return false;
	}

	if( !test_move_generation( "3k4/8/8/q2pP2K/8/8/8/8 w - d6",
						"e5-e6 Kh5-g4 Kh5-g5 Kh5-g6 Kh5-h4 Kh5-h6") )
	{
		return false;
	}

	return true;
}


static bool test_zobrist()
{
	position p;
	color::type c;
	if( !parse_fen_noclock( "rnbqk2r/1p3pp1/3bpn2/p2pN2p/P1Pp4/4P3/1P1BBPPP/RN1Q1RK1 b kq c3", p, c ) ) {
		return false;
	}

	uint64_t old_hash = get_zobrist_hash( p );

	move m;
	if( !parse_move( p, c, "dxc3", m ) ) {
		return false;
	}

	uint64_t new_hash_move = update_zobrist_hash( p, c, old_hash, m );

	apply_move( p, m, c );

	uint64_t new_hash_full = get_zobrist_hash( p );

	if( new_hash_move != new_hash_full ) {
		std::cerr << "Hash mismatch: " << new_hash_full << " " << new_hash_move << std::endl;
		return false;
	}

	return true;
}


static bool test_lazy_eval( std::string const& fen, short& max_difference )
{
	position p;
	color::type c;
	std::string error;
	if( !parse_fen_noclock( fen, p, c, &error ) ) {
		std::cerr << "Could not parse fen: " << error << std::endl;
		std::cerr << "Fen: " << fen << std::endl;
		return false;
	}

	score currents = c ? -p.base_eval : p.base_eval;
	short current = currents.scale( p.material[0].mg() + p.material[1].mg() );
	short full = evaluate_full( p, c );

	short diff = std::abs( full - current );

	if( diff > LAZY_EVAL ) {
		std::cerr << "Bug in lazy evaluation, full evaluation differs too much from basic evaluation:" << std::endl;
		std::cerr << "Position:   " << fen << std::endl;
		std::cerr << "Current:    " << current << std::endl;
		std::cerr << "Full:       " << full << std::endl;
		std::cerr << "Difference: " << diff << std::endl;
		return false;
	}

	max_difference = (std::max)( max_difference, diff );

	return true;
}


static bool test_lazy_eval()
{
	std::ifstream in_fen("test/testpositions.txt");

	short max_difference = 0;

	std::string fen;
	while( std::getline( in_fen, fen ) ) {
		if( !test_lazy_eval( fen, max_difference ) ) {
			return false;
		}
	}

	std::cout << "Max positional score difference: " << max_difference << std::endl;

	return true;
}


static bool test_pst() {
	for( int i = 0; i < 64; ++i ) {
		int opposite = (i % 8) + (7 - i / 8) * 8;
		int mirror = (i / 8) * 8 + 7 - i % 8;
		for( int p = 1; p < 7; ++p ) {
			if( pst[0][p][i] != pst[1][p][opposite] ) {
				std::cerr << "PST not symmetric for piece " << p << ", squares " << i << " and " << opposite << ": " << pst[0][p][i] << " " << pst[1][p][opposite] << std::endl;
				return false;
			}
			if( pst[0][p][i] != pst[0][p][mirror] ) {
				std::cerr << "PST not symmetric for piece " << p << ", squares " << i << " and " << mirror << ": " << pst[0][p][i] << " " << pst[0][p][mirror] << std::endl;
				return false;
			}
		}
	}

	return true;
}


static bool test_evaluation( std::string const& fen )
{
	position p;
	color::type c;
	std::string error;
	if( !parse_fen_noclock( fen, p, c, &error ) ) {
		std::cerr << "Could not parse fen: " << error << std::endl;
		std::cerr << "Fen: " << fen << std::endl;
		return false;
	}

	std::string flipped;

	std::string remaining;
	std::string first_part = split( fen, remaining );
	first_part += '/';
	while( !first_part.empty() ) {
		std::string segment = split( first_part, first_part, '/' );
		if( !flipped.empty() ) {
			flipped = '/' + flipped;
		}
		flipped = segment + flipped;
	}

	flipped += " ";
	if( remaining[0] == 'b' ) {
		flipped += "w";
	}
	else {
		flipped += "b";
	}

	if( remaining[remaining.size()] == '6' ) {
		remaining[remaining.size()] = '3';
	}
	else if( remaining[remaining.size()] == '3' ) {
		remaining[remaining.size()] = '6';
	}
	flipped += remaining.substr(1);

	for( std::size_t i = 0; i < flipped.size(); ++i ) {
		switch( flipped[i] ) {
		case 'p':
			flipped[i] = 'P';
			break;
		case 'n':
			flipped[i] = 'N';
			break;
		case 'b':
			flipped[i] = 'B';
			break;
		case 'r':
			flipped[i] = 'R';
			break;
		case 'q':
			flipped[i] = 'Q';
			break;
		case 'k':
			flipped[i] = 'K';
			break;
		case 'P':
			flipped[i] = 'p';
			break;
		case 'N':
			flipped[i] = 'n';
			break;
		case 'B':
			flipped[i] = 'b';
			break;
		case 'R':
			flipped[i] = 'r';
			break;
		case 'Q':
			flipped[i] = 'q';
			break;
		case 'K':
			flipped[i] = 'k';
			break;
		default:
			break;
		}
	}

	position p2;
	color::type c2;
	if( !parse_fen_noclock( flipped, p2, c2, &error ) ) {
		std::cerr << "Could not parse fen: " << error << std::endl;
		std::cerr << "Fen: " << flipped << std::endl;
		return false;
	}

	if( p.material[0] != p2.material[1] || p.material[1] != p2.material[0] ) {
		std::cerr << "Material not symmetric: " << p.material[0] << " " << p.material[1] << " " << p2.material[0] << " " << p2.material[1] << std::endl;
		std::cerr << "Fen: " << fen << std::endl;
		std::cerr << "Flipped: " << flipped << std::endl;
		return false;
	}

	if( p.base_eval.mg() != -p2.base_eval.mg() ||
		p.base_eval.eg() != -p2.base_eval.eg() )
	{
		std::cerr << "Base evaluation not symmetric: " << p.base_eval << " " << p2.base_eval << std::endl;
		std::cerr << "Fen: " << fen << std::endl;
		std::cerr << "Flipped: " << flipped << std::endl;
		return false;
	}

	short eval_full = evaluate_full( p, c );
	short flipped_eval_full = evaluate_full( p2, c2 );
	if( eval_full != flipped_eval_full ) {
		std::cerr << "Evaluation not symmetric: " << eval_full << " " << flipped_eval_full << " " << std::endl;
		std::cerr << "Fen: " << fen << std::endl;
		std::cerr << explain_eval( p, c ) << std::endl;
		std::cerr << "Flipped: " << flipped << std::endl;
		std::cerr << explain_eval( p2, c2 ) << std::endl;
		return false;
	}


	return true;
}


static bool test_evaluation()
{
	std::ifstream in_fen("test/testpositions.txt");

	std::string fen;
	while( std::getline( in_fen, fen ) ) {
		if( !test_evaluation( fen ) ) {
			return false;
		}
	}

	return true;
}


static bool do_selftest()
{
	if( !test_pst() ) {
		return false;
	}
	if( !test_evaluation() ) {
		return false;
	}
	if( !test_move_generation() ) {
		return false;
	}
	if( !test_zobrist() ) {
		return false;
	}
	if( !test_lazy_eval() ) {
		return false;
	}
	if( !perft(6) ) {
		return false;
	}

	return true;
}
}

bool selftest()
{
	pawn_hash_table.init( conf.pawn_hash_table_size );

	if( do_selftest() ) {
		std::cerr << "Self test passed" << std::endl;
		return true;
	}

	std::cerr << "Self test failed" << std::endl;
	abort();
	return false;
}
